from datetime import datetime
import os
from shutil import copy2, copytree
from pathlib import Path
from textwrap import dedent
import warnings
import sys
import yaml


from one_pass import __file__ as one_pass_file, __version__ as one_pass_version
from gsv import __file__ as gsv_file, __version__ as gsv_version

from opa_tools.config import get_benchmark_dir

BENCHMARK_DIR = get_benchmark_dir()


GIT_TREE = (
    "FETCH_HEAD",
    "HEAD",
    "ORIG_HEAD",
    "index",
    "refs",
)
DAY_DIGITS = 3


def copy_git_tree(init_file: str, dst: Path):
    dst.mkdir()
    git_directory = Path(init_file).parent.parent / ".git"
    if git_directory.exists():
        for rel_path in GIT_TREE:
            is_dir = (git_directory / rel_path).is_dir()
            copy_fun = copytree if is_dir else copy2
            dst_ = dst / rel_path if is_dir else dst
            copy_fun(git_directory / rel_path, dst_)
    else:
        with open(dst / "NO_GIT_DIRECTORY", "w"):
            pass


def create_benchmarking_path(now: datetime) -> Path:
    day_benchmark_path = BENCHMARK_DIR / now.strftime(r"%Y%m%d")
    day_benchmark_path.mkdir(parents=False, exist_ok=True)
    num_dirs = len(list(day_benchmark_path.iterdir()))
    if num_dirs == 0:
        next_benchmark_id = 1
    else:
        try:
            next_benchmark_id = max(
                int(path.name) for path in day_benchmark_path.iterdir()
            ) + 1
        except (ValueError, TypeError):
            warnings.warn(
                f"Malformed benchmarking directory: {day_benchmark_path}."
            )
            next_benchmark_id = num_dirs + 1
    benchmark_path = day_benchmark_path / f"{next_benchmark_id:0{DAY_DIGITS}d}"
    benchmark_path.mkdir(parents=False, exist_ok=False)
    return benchmark_path


def setup(*, environ=True, argv=True, **kwargs) -> Path:
    now = datetime.now()
    benchmark_dir = create_benchmarking_path(now)
    (benchmark_dir / "git").mkdir()
    copy_git_tree(gsv_file, benchmark_dir / "git" / "gsv")
    copy_git_tree(one_pass_file, benchmark_dir / "git" / "one_pass")

    with open(benchmark_dir / "version.txt", "w", encoding="utf-8") as wf:
        wf.write(
            dedent(
                f"""
            gsv: {gsv_version}
            one_pass: {one_pass_version}
            """
            )
        )

    with open(benchmark_dir / "date.txt", "w", encoding="utf-8") as wf:
        wf.write(now.isoformat())

    if environ:
        with open(benchmark_dir / "environ.yaml", "w", encoding="utf-8") as wf:
            yaml.dump(dict(os.environ), wf)

    if argv:
        with open(benchmark_dir / "argv.yaml", "w", encoding="utf-8") as wf:
            yaml.dump(sys.argv, wf)

    if kwargs:
        (benchmark_dir / "additional").mkdir()
        for key, value in kwargs.items():
            with open(
                benchmark_dir / "additional" / f"{key}.yaml", "w", encoding="utf-8"
            ) as wf:
                try:
                    yaml.dump(value, wf)
                except Exception as exc:
                    warnings.warn(f"Could not save additional data: '{key}'.")
                    continue

    return benchmark_dir
