from argparse import ArgumentParser
from functools import partial
from pandas import DataFrame
import sys
import warnings

from tabulate import tabulate

from opa_tools import get_request_df, get_environ_df
from opa_tools.utils import add_date_parsing_args, parse_date_interval

def _get_parser():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(title="Service", dest="service")

    parser_req = subparsers.add_parser("req")
    _add_table_arguments(parser_req)

    parser_env = subparsers.add_parser("env")
    _add_table_arguments(parser_env)

    parser_col = subparsers.add_parser("col")
    parser_col.add_argument("which", choices=("req", "env"), type=str, action="store", default="req")
    return parser

def _add_table_arguments(parser):
    parser.add_argument("-c", "--columns", action="store", nargs="*", default=[])
    add_date_parsing_args(parser)
    parser.add_argument(
        "-f", "--format",
        action="store",
        type=str, choices=["default", "fwf"] + [
            method.lstrip("to_")
            for method
            in dir(DataFrame)
            if method.startswith("to_")
        ],
        default="default",
    )
    parser.add_argument(
        "-t", "--tablefmt",
        action="store",
        type=str,
        metavar="TABULAR_TABLEFMT",
        default="plain",
    )

def _validate_columns(requested, existing):
    def force_column(lst, col):
        if col in existing:
            if col in lst:
                return
            lst.insert(0, col)
    columns = requested.copy()
    force_column(columns, "date")
    force_column(columns, "bname")
    wrong_columns = list(filter(lambda x: x not in existing, columns))
    if wrong_columns:
        warnings.warn(
            f"Column(s) {', '.join(wrong_columns)} skipped."
        )
        columns = [column for column in columns if column not in wrong_columns]
    return columns

def _table_service(args, fun):
    if not args.interval:
        df = fun(dates=args.dates)
    else:
        df = fun(date_interval=parse_date_interval(args.dates))
    if args.columns:
        columns = _validate_columns(args.columns, df.columns)
        df = df[columns]
    fmt = args.format
    if fmt == "default":
        print(df)
        return
    if fmt == "fwf":
        print(
            tabulate(
                df.values.tolist(),
                list(df.columns),
                tablefmt=args.tablefmt
            )
        )
        return
    if hasattr(df, f"to_{fmt}"):
        print(getattr(df, f"to_{fmt}")())
        return
    raise ValueError(
        f"Unknown format: '{fmt}'."
    )

def _column_service(args):
    df = (
        get_request_df()
        if args.which == "req"
        else get_environ_df()
    )
    print("\n".join(sorted(df.columns)))


_AVAILABLE_SERVICES = {
    "req": partial(_table_service, fun=get_request_df),
    "env": partial(_table_service, fun=get_environ_df),
    "col": partial(_column_service),
}

def bch(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = _get_parser()
    args = parser.parse_args(argv)
    service = args.service
    if not service:
        parser.print_help()
        sys.exit(1)
    _AVAILABLE_SERVICES[service](args)

