import ast
from argparse import ArgumentParser, Namespace
from collections import defaultdict
from datetime import datetime
from dataclasses import dataclass
import numpy as np
from pathlib import Path
from typing import Callable, Optional, Sequence, Tuple
import sys

from matplotlib.pyplot import cm
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.lines as mlines

from opa_tools.benchmarking.retrieval import get_benchmark_iterator, get_request_df
from opa_tools.benchmarking.processor import get_prof_time
from opa_tools.utils import (
    add_date_parsing_args,
    iterdict,
    parse_date_interval,
)


def _get_parser():
    parser = ArgumentParser()
    add_date_parsing_args(parser)
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        action="store",
        metavar="OUTPUT",
        default="$HOME/Pictures/world_coverage.png",
    )
    parser.add_argument(
        "-l", action="count", help="Legend", dest="legend", default=0
    )
    parser.add_argument(
        "--logarithmic", action="store_true", help="Make logarithmic y-axis"
    )
    parser.add_argument(
        "plot_nums",
        type=int,
        action="store",
        nargs="+",
    )
    return parser


def _get_benchmark_date(benchmark_dir):
    with open(benchmark_dir / "date.txt", "r") as rf:
        return rf.read()


def _get_color_iterator(cm_name=None, n=10):
    return iter(cm.viridis(np.linspace(0, 1, n)))
    # return cm.get_cmap(cm_name)


def _include_legend(ax, colors, markers=None, *, ncols=1, loc="best", **kwargs):
    fig_legend = []
    if markers is None:
        for txt_color, color in colors.items():
            fig_legend.append(mpatches.Patch(color=color, label=txt_color))
    else:
        for txt_color, color in colors.items():
            for txt_ls, marker in markers.items():
                fig_legend.append(
                    mlines.Line2D(
                        [],
                        [],
                        marker=marker,
                        color=color,
                        label=f"{txt_ls} - {txt_color}",
                    )
                )
    ax.legend(handles=fig_legend, loc=loc, ncols=ncols, **kwargs)


def plot_world_coverage(
    dates: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    interval: Optional[Tuple[datetime, datetime]] = None,
    logarithmic: bool = False,
    legend: bool = False,
    **kwargs,
):
    request_df = get_request_df(dates, exclude, interval)
    num_stats = len(set(request_df["opa_stat"]))
    colors = iterdict(_get_color_iterator(n=num_stats), cycle=False)
    print(request_df)
    fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    for benchmark_dir in get_benchmark_iterator(
        dates=dates, exclude=exclude, interval=interval
    ):
        benchmark_date = _get_benchmark_date(benchmark_dir)
        benchmark_name = benchmark_dir.name
        for file in benchmark_dir.iterdir():
            benchmark_req = request_df[
                (request_df["date"] == benchmark_date)
                & (request_df["bname"] == benchmark_name)
            ]
            opa_stat = benchmark_req["opa_stat"].iloc[0]
            gsv_area = ast.literal_eval(benchmark_req["gsv_area"].iloc[0])
            world_coverage = (
                (gsv_area[0] - gsv_area[2]) * (gsv_area[3] - gsv_area[1])
            ) / (180 * 360)
            if file.suffix == ".prof":
                # TODO: Remove hardcoded restrictions
                time = get_prof_time(file, "one_pass", "compute")
                color = colors[opa_stat]
                ax.plot([world_coverage], [time], "o", color=color, lw=0)
                ax.text(time, 0.5, benchmark_name)
                ax.text(time, -0.5, opa_stat)
                print(time)
    if legend:
        _include_legend(ax, colors)
    if logarithmic:
        ax.set_yscale("log")
    return fig


def plot_world_coverage_2(
    dates: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    interval: Optional[Tuple[datetime, datetime]] = None,
    logarithmic: bool = False,
    legend: int = 0,
    **kwargs,
):
    request_df = get_request_df(dates, exclude, interval)
    num_stats = len(set(request_df["opa_stat"]))
    colors = iterdict(_get_color_iterator(n=num_stats), cycle=False)
    markers = iterdict(["x", "v", "^"], cycle=False)
    print(request_df)
    fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    for benchmark_dir in get_benchmark_iterator(
        dates=dates, exclude=exclude, interval=interval
    ):
        benchmark_date = _get_benchmark_date(benchmark_dir)
        benchmark_name = benchmark_dir.name
        for file in benchmark_dir.iterdir():
            benchmark_req = (
                request_df[
                    (request_df["date"] == benchmark_date)
                    & (request_df["bname"] == benchmark_name)
                ]
                .iloc[0]
                .to_dict()
            )
            opa_stat = benchmark_req["opa_stat"]
            gsv_area = ast.literal_eval(benchmark_req["gsv_area"])
            world_coverage = (
                (gsv_area[0] - gsv_area[2]) * (gsv_area[3] - gsv_area[1])
            ) / (180 * 360)
            if file.suffix == ".prof":
                # TODO: Remove hardcoded restrictions
                time = get_prof_time(file, "one_pass", "compute")
                color = colors[opa_stat]
                marker = markers[(
                    benchmark_req["opa_checkpoint"],
                    benchmark_req["opa_save"],
                )]
                ax.plot([world_coverage], [time], marker=marker, color=color, lw=0, alpha=0.5)
                ax.text(time, 0.5, benchmark_name)
                ax.text(time, -0.5, opa_stat)
                print(time)
    print(legend)
    if legend == 0:
        pass
    elif legend == 1:
        _include_legend(ax, colors)
    elif legend == 2:
        _include_legend(ax, colors, markers, ncols=3, fontsize=8)
    else:
        raise ValueError(
            f"Wrong value for legend: {legend}."
        )
    if logarithmic:
        ax.set_yscale("log")
    return fig


@dataclass
class Plot:
    fun: Callable
    name: str

    def plot(self, args: Namespace):
        import seaborn as sns
        fun = self.fun
        dates = args.dates
        interval = args.interval
        logarithmic = args.logarithmic
        legend = args.legend
        figpath = Path.home() / "Pictures" / self.name
        kwargs = {"logarithmic": logarithmic, "legend": legend}
        fig = (
            fun(dates=dates, **kwargs)
            if not interval
            else fun(interval=parse_date_interval, **kwargs)
        )
        sns.set_theme()
        fig.savefig(figpath)
        print(f"Saved figure: {figpath}")
        return fig


PLOTS = {
    1: Plot(plot_world_coverage, "world_coverage"),
    2: Plot(plot_world_coverage_2, "world_coverage_2"),
}


def main():
    argv = sys.argv[1:]
    args = _get_parser().parse_args(argv)
    plot_nums = args.plot_nums
    for num in plot_nums:
        PLOTS[num].plot(args)


if __name__ == "__main__":
    main()
