"""Adapted from
https://earth.bsc.es/gitlab/digital-twins/de_340-2/workflow/-/blob/main/lib/runscript/run_opa.py
"""

from argparse import ArgumentParser
import cProfile
from functools import partial
import os
import sys

from pprint import pprint

from one_pass.opa import Opa
from opa_tools import get_request, benchmark_setup
from gsv import GSVRetriever

data = None
opa_stat = None
request = None
gsvrequest = None
oparequest = None
parrequest = None


def _get_parser():
    parser = ArgumentParser(description="Runscript for OPA job.")
    parser.add_argument(
        "-r", "--request", type=str, required=True, help="Input request number"
    )
    parser.add_argument(
        "--read_from_databridge", type=bool, help="Read from databridge", default=False
    )
    parser.add_argument(
        "-b",
        "--benchmark",
        action="store_true",
        help="Switch on benchmarking",
    )
    parser.add_argument(
        "-d",
        "--dummy",
        action="count",
        help=(
                "Stop before running. Useful for testing. "
                "-d prevents computation, "
                "-dd prevents data retrieval."
        ),
        default=0
    )
    return parser


def run_bias_adjustment(oparequest):  # type: ignore
    raise NotImplementedError(
        "OPA with bias adjustment not implemented in OPA-Tools package. "
        "Refer to the original script for implementation."
    )


def run(
    gsvrequest, oparequest, parrequest, *, benchmark, read_from_databridge, dummy=False
):
    global data, opa_stat

    profile = cProfile.Profile() if benchmark else False
    if profile:
        benchmark_path = benchmark_setup(
            environ=True, gsvrequest=gsvrequest, oparequest=oparequest, parrequest=parrequest
        )

    gsv = GSVRetriever()
    data = gsv.request_data(gsvrequest, use_stream_iterator=read_from_databridge)

    if data is None:
        raise SystemExit("No data was retreived with passed request pair.")

    if parrequest.get("forcemem", False):
        data.load()

    if dummy:
        raise KeyboardInterrupt(
            f"Interrupted by flag dummy={dummy}"
        )

    if oparequest["stat"] == "raw" and oparequest["bias_adjustment"] is True:
        run_bias_adjustment(oparequest)

    else:
        compute = (
            Opa.compute
            if not parrequest.get("parallel", False)
            else partial(
                Opa.compute_parallel,
                method=parrequest["method"],
                workers=parrequest["workers"],
            )
        )
        if profile:
            profile.enable()
        opa_stat = Opa(oparequest)
        compute(opa_stat, data)
        if profile:
            profile.disable()
            profile.dump_stats(benchmark_path / "opa.prof")  # type: ignore


def main(argv=None):
    global request, gsvrequest, oparequest, parrequest

    if argv is None:
        argv = sys.argv

    args = _get_parser().parse_args(argv[1:])
    request = args.request

    gsvrequest, oparequest, parrequest = get_request(request)
    print("\nGSV REQUEST:\n============\n")
    pprint(gsvrequest)
    print("\nOPA REQUEST:\n============\n")
    pprint(oparequest)

    if args.dummy >= 2:
        raise KeyboardInterrupt(
            f"Interrupted by flag dummy={args.dummy}"
        )

    run(
        gsvrequest,
        oparequest,
        parrequest=parrequest,
        benchmark=(args.benchmark and args.dummy == 0),
        read_from_databridge=args.read_from_databridge,
        dummy=bool(args.dummy),
    )


if __name__ == "__main__":
    if os.environ.get("FDB_HOME") is None:
        raise EnvironmentError("Please, set variable 'FDB_HOME'.")
    main()
